import asyncio
import hashlib
import logging
import os
import signal
import sys
import webbrowser
from multiprocessing import get_context
from typing import Text

from rasa.cli.utils import print_error, print_success
from rasa.data import get_core_nlu_files
from rasax.community import config, metrics, sql_migrations
from rasax.community.api.app import initialize_app
from rasax.community.api.initialise import (
    inject_config, inject_domain, inject_nlu_data, inject_stories, inject_user)
from rasax.community.database import utils
from rasax.community.server import configure_app
from rasax.community.services.data_service import DataService
from rasax.community.services.model_service import StackModelService
from rasax.community.services.nlg_service import NlgService
from rasax.community.services.projects_service import ProjectsService
from rasax.community.services.role_service import RoleService
from rasax.community.services.settings_service import SettingsService
from rasax.community.services.story_service import StoryService
from rasax.community.services.user_service import UserService
from rasax.community.utils import (
    accept_terms_or_quit, are_terms_accepted, is_enterprise_installed,
    metrics_collection_config, random_password, update_log_level)

logger = logging.getLogger(__name__)

LOCAL_PROJECT_NAME = "default"
LOCAL_TEAM_NAME = "rasa"
LOCAL_USERNAME = "me"
LOCAL_DATA_DIR = "data"
LOCAL_DEFAULT_NLU_FILENAME = "nlu.md"
LOCAL_DEFAULT_STORIES_FILENAME = "stories.md"
LOCAL_DOMAIN_PATH = "domain.yml"
LOCAL_CONFIG_PATH = "config.yml"
LOCAL_MODELS_DIR = "models"


def _configure_for_local_server(data_path, token):
    """Create `models` directory and set variables for local mode.

    Sets the API-wide token if provided.
    """

    if not os.path.isdir(LOCAL_MODELS_DIR):
        os.makedirs(LOCAL_MODELS_DIR)

    if token is not None:
        config.rasax_token = token

    config.data_dir = data_path
    config.stack_model_dir = LOCAL_MODELS_DIR
    config.project_name = LOCAL_PROJECT_NAME
    config.team_name = LOCAL_TEAM_NAME
    config.data_dir = LOCAL_DATA_DIR
    config.default_nlu_filename = LOCAL_DEFAULT_NLU_FILENAME
    config.default_stories_filename = LOCAL_DEFAULT_STORIES_FILENAME
    config.default_username = LOCAL_USERNAME


def update_local_password(settings_service: "SettingsService") -> Text:
    """Update Rasa X password.

    Check if `RASAX_PASSWORD` environment variable is set, and use it as password.
    Generate new password if it is not set.
    """

    password = os.environ.get("RASAX_PASSWORD", random_password())

    # update password in config
    config.rasax_password = password

    # commit password to db
    settings_service.save_local_password(password)

    return password


def _check_license_and_metrics(args):
    if not are_terms_accepted():
        accept_terms_or_quit(args)

    config.metrics_collection_config = metrics_collection_config(args)


def _enable_development_mode(app):
    if config.development_mode:
        if not is_enterprise_installed():
            print_error(
                "Rasa X EE is not installed. Using enterprise endpoints in "
                "local development mode requires an installation of "
                "Rasa X Enterprise Edition."
            )
            sys.exit(1)

        from rasax.enterprise.server import configure_enterprise_endpoints
        configure_enterprise_endpoints(app)


def _event_service():
    """Start the event service."""
    # noinspection PyUnresolvedReferences
    from rasax.community.services import event_service

    def signal_handler(sig, frame):
        print("Stopping event service.")
        os.kill(os.getpid(), signal.SIGTERM)

    signal.signal(signal.SIGINT, signal_handler)
    event_service.main()


def start_event_service():
    """Run the event service in a separate process."""

    ctx = get_context("spawn")
    p = ctx.Process(target=_event_service)
    p.start()


def _initialize_with_local_data(project_path, data_path, session):
    user_service = UserService(session)
    role_service = RoleService(session)
    settings_service = SettingsService(session)

    password = update_local_password(settings_service)

    inject_user(
        user_service, role_service, LOCAL_USERNAME, password,
        LOCAL_PROJECT_NAME, LOCAL_TEAM_NAME
    )

    projects_service = ProjectsService(session)
    nlg_service = NlgService(session)
    domain = inject_domain(
        os.path.join(project_path, LOCAL_DOMAIN_PATH),
        projects_service,
        nlg_service,
        LOCAL_PROJECT_NAME,
    )
    settings_service = SettingsService(session)
    inject_config(os.path.join(project_path, LOCAL_CONFIG_PATH), settings_service)

    # discover model
    model_service = StackModelService(LOCAL_MODELS_DIR,
                                      session, "production")

    model_service.discover_models_on_init()
    model_service.mark_latest_as_production()

    session.commit()
    session.close()

    story_files, nlu_files = get_core_nlu_files([data_path])

    loop = asyncio.get_event_loop()
    story_service = StoryService(session)
    story_blocks = loop.run_until_complete(
        inject_stories(story_files, story_service, LOCAL_USERNAME, LOCAL_TEAM_NAME)
    )

    data_service = DataService(session)
    nlu_data = inject_nlu_data(nlu_files, LOCAL_PROJECT_NAME,
                               LOCAL_USERNAME, data_service)

    session.commit()
    session.close()

    return password, domain, story_blocks, nlu_data


def main(
        args, project_path: Text, data_path: Text, token: Text = None
):
    """Start Rasa X in local mode."""

    _check_license_and_metrics(args)

    print_success("Starting Rasa X in local mode... 🚀")

    _configure_for_local_server(data_path, token)

    session = utils.get_database_session(config.LOCAL_MODE, create_tables=True)
    app = configure_app()
    _enable_development_mode(app)
    initialize_app(app)
    sql_migrations.run_migrations(session)

    password, domain, story_blocks, nlu_data = _initialize_with_local_data(project_path,
                                                                           data_path,
                                                                           session)

    # this needs to run after initial database structures are created
    # otherwise projects assigned to events won't be present
    start_event_service()

    async def startup_info(app, loop):
        """Open a browser window pointing to the UI."""

        if config.rasax_hostname:
            server_url = config.rasax_hostname
        else:
            server_url = "http://localhost:{}".format(config.self_port)

        login_url = ("{}/login?username={}&password={}"
                     "".format(server_url, LOCAL_USERNAME, password))

        print_success("\nThe server is running at {}\n".format(login_url))
        webbrowser.open_new_tab(login_url)

    app.register_listener(startup_info, "after_server_start")

    update_log_level()

    metrics.track("Local X Start", {
        "project": hashlib.md5(project_path.encode('utf-8')).hexdigest(),
        "num_intent_examples": len(nlu_data.intent_examples),
        "num_entity_examples": len(nlu_data.entity_examples),
        "num_actions": len(domain.get("actions", [])),
        "num_templates": len(domain.get("templates", [])),
        "num_slots": len(domain.get("slots", [])),
        "num_forms": len(domain.get("forms", [])),
        "num_intents": len(domain.get("intents", [])),
        "num_entities": len(domain.get("entities", [])),
        "num_stories": len(story_blocks),
    })

    app.run(
        host="0.0.0.0",
        port=config.self_port,
        auto_reload=os.environ.get("SANIC_AUTO_RELOAD"),
        access_log=False,
    )
